/*
 * rxv_spin.h
 *
 * Copyright (C) 2003 - 2007 Bojan Smojver, Rexursive
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public Licence as published by the Free
 * Software Foundation; either version 2 of the Licence, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public Licence for
 * more details.
 *
 * You should have received a copy of the GNU General Public Licence along
 * with this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * In addition, as two special exceptions, Bojan Smojver, Rexursive, gives
 * permission to:
 *
 * 1. Link, both statically and dynamically, the code of this program with
 * Apache HTTP Server, Apache Portable Runtime and Apache Portable Runtime
 * Utility Library from Apache Software Foundation (or with modified versions
 * of the above, that use the same licence - Apache Software Licence 1.1, 2.0
 * or any later version), and distribute linked combinations including the
 * program and the above software from Apache Software Foundation. You must
 * obey the GNU General Public Licence in all respects for all of the code
 * used other than Apache HTTP Server, Apache Portable Runtime and Apache
 * Portable Runtime Utility Library.
 *
 * 2. Dynamically link any shared library with this program at run-time,
 * through the interface of SpinApplication/SpinAppEntry or LoadModule
 * run-time configuration directives of Apache HTTP Server, as provided by
 * this program or Apache HTTP Server itself, regardless of licensing terms of
 * those shared libraries. You must obey the GNU General Public Licence in all
 * respects for all of the code used other than that of those shared
 * libraries.
 *
 * If you modify this file, you may extend these exceptions to your version of
 * the file, but you are not obligated to do so. If you do not wish to do so,
 * delete one or both of these exception statements from your version.
 *
 */

#ifndef __RXV_SPIN__
#define __RXV_SPIN__

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <apr.h>
#include <apr_general.h>
#include <apr_strings.h>
#include <apr_lib.h>
#include <apr_pools.h>
#include <apr_buckets.h>
#include <apr_hash.h>
#include <apr_file_io.h>
#include <apr_portable.h>
#include <apr_time.h>
#include <apr_sdbm.h>
#include <apr_md5.h>
#include <apr_base64.h>
#include <apreq_param.h>
#include <apreq_cookie.h>

/* remove some macros, Apache defines them as well */
#ifndef DOXYGEN_SHOULD_SKIP_THIS
#undef  PACKAGE_BUGREPORT
#undef  PACKAGE_NAME
#undef  PACKAGE_STRING
#undef  PACKAGE_TARNAME
#undef  PACKAGE_VERSION
#endif

/* the following also includes httpd.h */
#include <apreq_module_apache2.h>

/* remove Apache macros, restore good definitions */
#ifndef DOXYGEN_SHOULD_SKIP_THIS
#undef  PACKAGE_BUGREPORT
#undef  PACKAGE_NAME
#undef  PACKAGE_STRING
#undef  PACKAGE_TARNAME
#undef  PACKAGE_VERSION
#define PACKAGE_BUGREPORT RXV_SPIN_PACKAGE_BUGREPORT
#define PACKAGE_NAME      RXV_SPIN_PACKAGE_NAME
#define PACKAGE_STRING    RXV_SPIN_PACKAGE_STRING
#define PACKAGE_TARNAME   RXV_SPIN_PACKAGE_TARNAME
#define PACKAGE_VERSION   RXV_SPIN_PACKAGE_VERSION
#endif

#include <http_request.h>

#ifndef DOXYGEN_SHOULD_SKIP_THIS
#ifdef HAVE_AP_REGEX_H
#include <ap_regex.h>
#else
#define ap_regex_t        regex_t
#define ap_regmatch_t     regmatch_t
#define ap_regcomp(a,b,c) regcomp((a),(b),(c))
#define ap_regfree(a)     regfree(a)
#define AP_REG_ICASE      REG_ICASE
#define AP_REG_NEWLINE    REG_NEWLINE
#define AP_REG_NOTBOL     REG_NOTBOL
#define AP_REG_NOTEOL     REG_NOTEOL
#define AP_REG_EXTENDED   REG_EXTENDED
#define AP_REG_NOSUB      REG_NOSUB
#endif
#endif

#include <libxml/parser.h>
#include <libxml/tree.h>

#ifdef HAVE_LIBPQ
#include <libpq-fe.h>
#endif
#ifdef HAVE_LIBMYSQLCLIENT
#include <mysql.h>
#endif

#define RXV_SPIN_DATA_SGL 0x01
  
#define RXV_SPIN_DATA_RWS 0x02
  
#define RXV_SPIN_DATA_MTA 0xFF
  

#define RXV_SPIN_TRIM_LEFT  0x01
  
#define RXV_SPIN_TRIM_RIGHT 0x02
  

typedef struct rxv_spin_data rxv_spin_data_t; 

typedef struct rxv_spin_context   rxv_spin_context_t; 
typedef        rxv_spin_context_t rxv_spin_ctx_t;     

typedef struct rxv_spin_conn  rxv_spin_conn_t;  
typedef struct rxv_spin_cpool rxv_spin_cpool_t; 

typedef struct rxv_spin_db_result rxv_spin_db_result_t; 

/* Data and context structures */

struct rxv_spin_data{
  unsigned char type;  
  size_t        size;  
  
  union{
    apr_hash_t      *cols; 
    char            *data; 
    rxv_spin_data_t *meta; 
    void            *both; 
  };
};

struct rxv_spin_context{
  request_rec      *r;     
  apr_pool_t       *pool;  
  rxv_spin_data_t  *data;  
  apreq_handle_t   *req;   
  rxv_spin_cpool_t *cpool; 
  void             *guts;  
  void             *extra; 
};

/* Data creation and manipulation functions */

rxv_spin_data_t *rxv_spin_single(apr_pool_t *pool,const char *str);

char *rxv_spin_single_get(rxv_spin_data_t *single);

rxv_spin_data_t *rxv_spin_single_set(rxv_spin_data_t *single,const char *str);

rxv_spin_data_t *rxv_spin_single_mem(apr_pool_t *pool,const char *str,
                                     size_t size);

rxv_spin_data_t *rxv_spin_single_memset(rxv_spin_data_t *single,const char *str,
                                        size_t size);

rxv_spin_data_t *rxv_spin_single_tolower(rxv_spin_data_t *single);

rxv_spin_data_t *rxv_spin_single_toupper(rxv_spin_data_t *single);

rxv_spin_data_t *rxv_spin_single_trim(rxv_spin_data_t *single,
                                      unsigned char what);
#define rxv_spin_single_trimboth(s) \
        rxv_spin_single_trim((s),(RXV_SPIN_TRIM_LEFT|RXV_SPIN_TRIM_RIGHT))
  
#define rxv_spin_single_trimleft(s) \
        rxv_spin_single_trim((s),(RXV_SPIN_TRIM_LEFT))
  
#define rxv_spin_single_trimright(s) \
        rxv_spin_single_trim((s),(RXV_SPIN_TRIM_RIGHT))
  

rxv_spin_data_t *rxv_spin_meta(apr_pool_t *pool,...);

rxv_spin_data_t *rxv_spin_meta_vstr(apr_pool_t *pool,...);

rxv_spin_data_t *rxv_spin_meta_parse(apr_pool_t *pool,
                                     char *str,const char *sep);

rxv_spin_data_t *rxv_spin_meta_empty(apr_pool_t *pool,size_t size);

apr_hash_t *rxv_spin_meta_hash(apr_pool_t *pool,rxv_spin_data_t *data);

rxv_spin_data_t *rxv_spin_meta_mark(rxv_spin_data_t *data,size_t element);

rxv_spin_data_t *rxv_spin_meta_markeach(rxv_spin_data_t *data,
                                        size_t off,size_t step);

rxv_spin_data_t *rxv_spin_meta_select(rxv_spin_data_t *data,
                                      rxv_spin_data_t *select,
                                      apr_hash_t *hash);

rxv_spin_data_t *rxv_spin_rows(apr_pool_t *pool,...);

apr_hash_t *rxv_spin_rows_hash(apr_pool_t *pool,rxv_spin_data_t *rows,
                               const char *column);

rxv_spin_data_t *rxv_spin_rows_mark(rxv_spin_data_t *rows,const char *column,
                                    size_t element);

rxv_spin_data_t *rxv_spin_rows_markeach(rxv_spin_data_t *rows,
                                        const char *column,
                                        size_t off,size_t step);

rxv_spin_data_t *rxv_spin_rows_select(rxv_spin_data_t *rows,
                                      rxv_spin_data_t *select,
                                      const char *column,const char *marker,
                                      apr_hash_t *hash);

rxv_spin_data_t *rxv_spin_column_get(apr_pool_t *pool,
                                     rxv_spin_data_t *rows,const char *key);

rxv_spin_data_t *rxv_spin_column_set(rxv_spin_data_t *rows,
                                     const char *key,rxv_spin_data_t *column);
#define rxv_spin_column_del(rows,key) rxv_spin_column_set((rows),(key),NULL)
  

rxv_spin_data_t *rxv_spin_resize(apr_pool_t *pool,rxv_spin_data_t *data,
                                 size_t size);

rxv_spin_data_t *rxv_spin_copy(apr_pool_t *pool,rxv_spin_data_t *data);

char *rxv_spin_str_tolower(const char *str);

char *rxv_spin_str_toupper(const char *str);

char *rxv_spin_str_trim(char *str,unsigned char what);
#define rxv_spin_str_trimboth(s) \
        rxv_spin_str_trim((s),(RXV_SPIN_TRIM_LEFT|RXV_SPIN_TRIM_RIGHT))
  
#define rxv_spin_str_trimleft(s) \
        rxv_spin_str_trim((s),(RXV_SPIN_TRIM_LEFT))
  
#define rxv_spin_str_trimright(s) \
        rxv_spin_str_trim((s),(RXV_SPIN_TRIM_RIGHT))
  

/* Context functions and macros */

rxv_spin_data_t *rxv_spin_ctx_get(rxv_spin_context_t *ctx,const char *key);
#define rxv_spin_ctx_strget(ctx,key) \
        rxv_spin_single_get(rxv_spin_ctx_get((ctx),(key)))
  

rxv_spin_data_t *rxv_spin_ctx_set(rxv_spin_context_t *ctx,
                                  const char *key,rxv_spin_data_t *value);
#define rxv_spin_ctx_strset(ctx,key,val) \
        rxv_spin_ctx_set((ctx),(key),rxv_spin_single((ctx)->pool,(val)))
  
#define rxv_spin_ctx_del(ctx,key) rxv_spin_ctx_set((ctx),(key),NULL)
  

/* Application/session functions */

rxv_spin_data_t *rxv_spin_app_get(rxv_spin_context_t *ctx,const char *key);
#define rxv_spin_app_strget(ctx,key) \
        rxv_spin_single_get(rxv_spin_app_get((ctx),(key)))
  

rxv_spin_data_t *rxv_spin_app_set(rxv_spin_context_t *ctx,
                                  const char *key,rxv_spin_data_t *val);
#define rxv_spin_app_strset(ctx,key,val) \
        rxv_spin_app_set((ctx),(key),rxv_spin_single((ctx)->pool,(val)))
  

apr_status_t rxv_spin_app_del(rxv_spin_context_t *ctx,const char *key);

rxv_spin_data_t *rxv_spin_ses_get(rxv_spin_context_t *ctx,const char *key);
#define rxv_spin_ses_strget(ctx,key) \
        rxv_spin_single_get(rxv_spin_ses_get((ctx),(key)))
  

rxv_spin_data_t *rxv_spin_ses_set(rxv_spin_context_t *ctx,
                                  const char *key,rxv_spin_data_t *val);
#define rxv_spin_ses_strset(ctx,key,val) \
        rxv_spin_ses_set((ctx),(key),rxv_spin_single((ctx)->pool,(val)))
  

apr_status_t rxv_spin_ses_del(rxv_spin_context_t *ctx,const char *key);

char *rxv_spin_ses_idget(rxv_spin_context_t *ctx);

int rxv_spin_ses_valid(rxv_spin_context_t *ctx);

/* Connection macros */

#define RXV_SPIN_CONN_PGSQL   0x01
  
#define RXV_SPIN_CONN_MYSQL   0x02
  
#define RXV_SPIN_CONN_MINID   0x01
  
#define RXV_SPIN_CONN_MAXID   0x3F /* that's six ones in binary */
  
#define RXV_SPIN_CONN_FOREIGN 0x40
  
#define RXV_SPIN_CONN_POOLED  0x80
  

#define RXV_SPIN_CONN_IS_PGSQL(c)   ((((c)->type)&RXV_SPIN_CONN_MAXID)==RXV_SPIN_CONN_PGSQL)
  
#define RXV_SPIN_CONN_IS_MYSQL(c)   ((((c)->type)&RXV_SPIN_CONN_MAXID)==RXV_SPIN_CONN_MYSQL)
  
#define RXV_SPIN_CONN_IS_FOREIGN(c) (((c)->type)&RXV_SPIN_CONN_FOREIGN)
  
#define RXV_SPIN_CONN_IS_POOLED(c)  (((c)->type)&RXV_SPIN_CONN_POOLED)
  

struct rxv_spin_conn{ 
  unsigned char type;   
  char          *cinfo; 
  
  union{
#ifdef HAVE_LIBPQ
    PGconn *pgconn; 
#endif
#ifdef HAVE_LIBMYSQLCLIENT
    MYSQL  *myconn; 
#endif
    void *conn;     
  };
  
  union{
    rxv_spin_cpool_t *cpool; 
    apr_pool_t       *pool;  
  };
  apr_status_t (*cleanup)(void *data); 
};

struct rxv_spin_cpool{ 
  apr_pool_t *pool;  
  apr_hash_t *conns; 
};

/* Database access structures and functions
   NOTE: Everything is designed to be PostgreSQL centric. Sorry :-( */

struct rxv_spin_db_result{ 
  apr_status_t    status; 
  char            *error; 
  rxv_spin_data_t *data;  
};

rxv_spin_conn_t *rxv_spin_db_connect(apr_pool_t *pool,
                                     rxv_spin_cpool_t *cpool,
                                     const char *conninfo,
                                     unsigned char type);

rxv_spin_db_result_t *rxv_spin_db_exec(apr_pool_t *pool,
                                       rxv_spin_conn_t *conn,
                                       const char *query);

char *rxv_spin_db_info(rxv_spin_conn_t *conn,unsigned char what);

apr_status_t rxv_spin_db_status(rxv_spin_conn_t *conn);

char *rxv_spin_db_escape(apr_pool_t *pool,
                         rxv_spin_conn_t *conn,
                         const char *str);

apr_status_t rxv_spin_db_reset(rxv_spin_conn_t *conn);

rxv_spin_cpool_t *rxv_spin_cpool_create(apr_pool_t *pool);

rxv_spin_conn_t *rxv_spin_cpool_get(apr_pool_t *pool,rxv_spin_cpool_t *cpool,
                                    const char *conninfo);

rxv_spin_conn_t *rxv_spin_cpool_set(rxv_spin_cpool_t *cpool,
                                    const char *conninfo,void *conn,
                                    apr_status_t (*cleanup)(void *data));

apr_status_t rxv_spin_conn_close(rxv_spin_conn_t *conn);

/* Database info flags */

#define RXV_SPIN_DB_INFO_DB      0x01
  
#define RXV_SPIN_DB_INFO_USER    0x02
  
#define RXV_SPIN_DB_INFO_PASS    0x03
  
#define RXV_SPIN_DB_INFO_HOST    0x04
  
#define RXV_SPIN_DB_INFO_PORT    0x05
  
#define RXV_SPIN_DB_INFO_TTY     0x06
  
#define RXV_SPIN_DB_INFO_OPTIONS 0x07
  
#define RXV_SPIN_DB_INFO_ERROR   0x08
  

/* Database info macros */

#define rxv_spin_db_db(conn)      \
        rxv_spin_db_info((conn),RXV_SPIN_DB_INFO_DB)
  
#define rxv_spin_db_user(conn)    \
        rxv_spin_db_info((conn),RXV_SPIN_DB_INFO_USER)
  
#define rxv_spin_db_pass(conn)    \
        rxv_spin_db_info((conn),RXV_SPIN_DB_INFO_PASS)
  
#define rxv_spin_db_host(conn)    \
        rxv_spin_db_info((conn),RXV_SPIN_DB_INFO_HOST)
  
#define rxv_spin_db_port(conn)    \
        rxv_spin_db_info((conn),RXV_SPIN_DB_INFO_PORT)
  
#define rxv_spin_db_tty(conn)     \
        rxv_spin_db_info((conn),RXV_SPIN_DB_INFO_TTY)
  
#define rxv_spin_db_options(conn) \
        rxv_spin_db_info((conn),RXV_SPIN_DB_INFO_OPTIONS)
  
#define rxv_spin_db_error(conn)   \
        rxv_spin_db_info((conn),RXV_SPIN_DB_INFO_ERROR)
  

typedef int (*rxv_spin_service_t)(rxv_spin_context_t *context);
  

/* Legacy macros for backward compatibility - MAY BE REMOVED ANYTIME! */

#ifndef DOXYGEN_SHOULD_SKIP_THIS

#define rxv_spin_str_tometa(p,...)    rxv_spin_meta_vstr((p),...)
#define rxv_spin_column(p,s)          rxv_spin_meta_empty((p),(s))
#define rxv_spin_column_mark(c,e)     rxv_spin_meta_mark((c),(e))
#define rxv_spin_column_markeach(c,n) rxv_spin_meta_markeach((c),0,(n))
#define rxv_spin_single_tostr(p,s)    rxv_spin_single_get(s)
#define rxv_spin_ctx_str_set(c,k,v)   rxv_spin_ctx_strset((c),(k),(v))
#define rxv_spin_app_str_get(c,k)     rxv_spin_app_strget((c),(k))
#define rxv_spin_app_str_set(c,k,v)   rxv_spin_app_strset((c),(k),(v))
#define rxv_spin_ses_str_get(c,k)     rxv_spin_ses_strget((c),(k))
#define rxv_spin_ses_str_set(c,k,v)   rxv_spin_ses_strset((c),(k),(v))
#define rxv_spin_db_conn              rxv_spin_conn
#define rxv_spin_db_conn_t            rxv_spin_conn_t
#define rxv_spin_db_pool              rxv_spin_cpool
#define rxv_spin_db_pool_t            rxv_spin_cpool_t
#define rxv_spin_db_pool_create(p)    rxv_spin_cpool_create(p)
#define rxv_spin_db_finish(c)         rxv_spin_conn_close(c)
#define RXV_SPIN_DB_PGSQL             RXV_SPIN_CONN_PGSQL
#define RXV_SPIN_DB_MYSQL             RXV_SPIN_CONN_MYSQL
#define RXV_SPIN_DB_MINID             RXV_SPIN_CONN_MINID
#define RXV_SPIN_DB_MAXID             RXV_SPIN_CONN_MAXID
#define RXV_SPIN_DB_POOLED            RXV_SPIN_CONN_POOLED
#define RXV_SPIN_DB_IS_PGSQL(c)       RXV_SPIN_CONN_IS_PGSQL(c)
#define RXV_SPIN_DB_IS_MYSQL(c)       RXV_SPIN_CONN_IS_MYSQL(c)
#define RXV_SPIN_DB_IS_POOLED(c)      RXV_SPIN_CONN_IS_POOLED(c)

#endif

/* End of legacy macros */

#endif
